import sys
from argparse import ArgumentParser
from time import sleep
from typing import Union
from dreamberd.base import InterpretationError, Token, TokenType

from dreamberd.builtin import KEYWORDS, Name, Variable
from dreamberd.processor.lexer import tokenize
from dreamberd.processor.syntax_tree import CodeStatement, generate_syntax_tree
from dreamberd.interpreter import interpret_code_statements, load_global_dreamberd_variables, load_globals, load_public_global_variables

__all__ = ['run_repl', 'run_file']

__REPL_FILENAME = "__repl__"
sys.setrecursionlimit(100000)

def __get_next_repl_input(closed_scope_layers: int = 0) -> tuple[str, list[Token]]:
    print("   " * closed_scope_layers, '\033[33m>\033[39m ', end="", sep="")
    code = input()
    tokens = tokenize(__REPL_FILENAME, code)
    start_closed_scope_layers = closed_scope_layers
    for t in tokens:
        if t.type == TokenType.L_CURLY:
            closed_scope_layers += 1
        elif t.type == TokenType.R_CURLY:
            closed_scope_layers -= 1
    if closed_scope_layers < 0:
        raise InterpretationError("Too many closed braces with not enough open braces.")
    kickback_layers, i = 0, 0
    trimmed_code = code.strip()
    while i < len(trimmed_code) and trimmed_code[i] == '}':
        kickback_layers += 1
        i += 1
    if kickback_layers:
        print("\033[1A\r", "   " * (start_closed_scope_layers - kickback_layers), '\033[33m>\033[39m ', code, "   " * (start_closed_scope_layers), sep="", flush=True)
    if closed_scope_layers:
        new_code, new_tokens = __get_next_repl_input(closed_scope_layers) 
        code += new_code 
        tokens += new_tokens
    match l := [t for t in tokens if not t.type == TokenType.WHITESPACE]:
        case [Token(type=TokenType.NAME)]:
            tokens = [Token(TokenType.NAME, "print", -1, -1), Token(TokenType.WHITESPACE, " ", -1, -1), l[0], Token(TokenType.BANG, '!', -1, -1)]
    return code, tokens

def run_repl() -> None:
    namespaces: list[dict[str, Union[Variable, Name]]] = [KEYWORDS.copy()]  # type: ignore
    load_globals(__REPL_FILENAME, "", {}, set())
    load_global_dreamberd_variables(namespaces)
    async_statements = []
    when_statement_watchers = [{}]
    while True: 
        code, tokens = __get_next_repl_input()
        statements = generate_syntax_tree(__REPL_FILENAME, tokens, code)
        interpret_code_statements(statements, namespaces, async_statements, when_statement_watchers)
            
def run_file(filename: str) -> None:  # idk what else to call this

    with open(filename, 'r') as f:
        code = f.read()
    tokens = tokenize(filename, code)
    statements = generate_syntax_tree(filename, tokens, code)

    namespaces: list[dict[str, Union[Variable, Name]]] = [KEYWORDS.copy()]   # type: ignore
    load_globals(filename, code, {}, set())
    load_global_dreamberd_variables(namespaces)
    load_public_global_variables(namespaces)
    interpret_code_statements(statements, namespaces, [], [{}])
    print("\033[33mCode has finished executing. Press ^C once or twice to stop waiting for when-statements and after-statements.\033[039m")
    try:
        while True:
            sleep(1)  # just waiting for any clicks, when statements, etc
    except KeyboardInterrupt:
        exit()  # quit silently

def parse_args():
    parser = ArgumentParser()
    parser.add_argument('file', help="The file containing your DreamBerd code.", nargs='?', default='', type=str)
    parser.add_argument('-s', '--suppress-traceback', help="Limit the error traceback to a single message.", action="store_true")
    return parser.parse_args()

def main():
    args = parse_args()
    if args.suppress_traceback:
        sys.tracebacklimit = 0
    if not args.file:
        run_repl()
    else:
        run_file(args.file)

if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Optional 

from dreamberd.base import Token, TokenType, ALPH_NUMS, raise_error_at_line

# thanks : https://craftinginterpreters.com/scanning.html

def add_to_tokens(token_list: list[Token], line: int, col: int, token: TokenType, value: Optional[str] = None):
    token_list.append(Token(token, value if value is not None else token.value, line, col))

def tokenize(filename: str, code: str) -> list[Token]:
    code += '   '  # adding a space here so i dont have to write 10 damn checks for out of bounds
    line_count = 1
    tokens = []
    curr, start = 0, 0
    while curr < len(code):
        match code[curr]:
            case '\n':
                line_count += 1
                start = curr  # at the new line to get col number
                add_to_tokens(tokens, line_count, curr - start, TokenType.NEWLINE)
            case '}': add_to_tokens(tokens, line_count, curr - start, TokenType.R_CURLY)
            case '{': add_to_tokens(tokens, line_count, curr - start, TokenType.L_CURLY)
            case '[': add_to_tokens(tokens, line_count, curr - start, TokenType.L_SQUARE)
            case ']': add_to_tokens(tokens, line_count, curr - start, TokenType.R_SQUARE)
            case '.': add_to_tokens(tokens, line_count, curr - start, TokenType.DOT)
            case ':': add_to_tokens(tokens, line_count, curr - start, TokenType.COLON)
            case '|': add_to_tokens(tokens, line_count, curr - start, TokenType.PIPE)
            case '&': add_to_tokens(tokens, line_count, curr - start, TokenType.AND)
            case ';': 
                value = ';'
                while code[curr + 1] == '=':
                    value += '='
                    curr += 1
                if len(value) == 1:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.SEMICOLON)
                else:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.NOT_EQUAL, value)
            case ',': add_to_tokens(tokens, line_count, curr - start, TokenType.COMMA)
            case '+': 
                if code[curr + 1] == '+':
                    add_to_tokens(tokens, line_count, curr - start, TokenType.INCREMENT)
                    curr += 1
                else:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.ADD)  # YOU NEVER SAID I HAD TO DO +=
            case '-': 
                if code[curr + 1] == '-':
                    add_to_tokens(tokens, line_count, curr - start, TokenType.DECREMENT)
                    curr += 1
                else:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.SUBTRACT)
            case '*': add_to_tokens(tokens, line_count, curr - start, TokenType.MULTIPLY)
            case '/': add_to_tokens(tokens, line_count, curr - start, TokenType.DIVIDE)
            case '^': add_to_tokens(tokens, line_count, curr - start, TokenType.CARROT)
            case '>': 
                if code[curr + 1] == '=':
                    add_to_tokens(tokens, line_count, curr - start, TokenType.GREATER_EQUAL)
                    curr += 1
                else:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.GREATER_THAN)
            case '<': 
                if code[curr + 1] == '=':
                    add_to_tokens(tokens, line_count, curr - start, TokenType.LESS_EQUAL)
                    curr += 1
                else:
                    add_to_tokens(tokens, line_count, curr - start, TokenType.LESS_THAN)
            case '!':
                value = '!'
                while code[curr + 1] == '!':
                    value += '!'
                    curr += 1
                add_to_tokens(tokens, line_count, curr - start, TokenType.BANG, value)
            case '?': 
                value = '?'
                while code[curr + 1] == '?':
                    value += '?'
                    curr += 1
                if len(value) > 4:
                    raise_error_at_line(filename, code, line_count, "User is too confused. Aborting due to trust issues.")  # heheheheheheh
                add_to_tokens(tokens, line_count, curr - start, TokenType.QUESTION, value)
            case '=': 
                value = '='
                if code[curr + 1] == '>':
                    curr += 1 
                    add_to_tokens(tokens, line_count, curr - start, TokenType.FUNC_POINT)
                else: 
                    while code[curr + 1] == '=':
                        value += '='
                        curr += 1
                    add_to_tokens(tokens, line_count, curr - start, TokenType.EQUAL, value)
            case '"' | "'":
                quote_count = 0
                while code[curr] in "'\"":
                    quote_count += 1 if code[curr] == "'" else 2
                    curr += 1
                value = ''
                while code[curr] not in "'\"\n":   # changing this here because \n not allowed in strings
                    value += code[curr]
                    if code[curr] == "\\":
                        curr += 1
                        value += code[curr]
                    curr += 1
                while code[curr] in "'\"":
                    quote_count -= 1 if code[curr] == "'" else 2
                    curr += 1
                if code[curr] not in '!\n':  # autocomplete of strings.......................
                    if quote_count != 0:
                        raise_error_at_line(filename, code, line_count, "Invalid string. Starting quotes do not match opening quotes.")
                curr -= 1
                add_to_tokens(tokens, line_count, curr - start, TokenType.STRING, value)
            case ' ' | '\t' | '(' | ')': 
                if code[curr] == '(' and curr + 1 < len(code) and code[curr + 1] == ')':
                    add_to_tokens(tokens, line_count, curr - start, TokenType.WHITESPACE, '')
                    add_to_tokens(tokens, line_count, curr - start, TokenType.NAME, '')  # please please please work
                    add_to_tokens(tokens, line_count, curr - start, TokenType.WHITESPACE, '')
                    curr += 1
                else:
                    value = code[curr] if code[curr] not in '()' else ' '
                    while curr + 1 < len(code) and code[curr + 1] in ' ()\t':
                        if code[curr + 1] not in "()":
                            value += code[curr + 1]
                        curr += 1
                    add_to_tokens(tokens, line_count, curr - start, TokenType.WHITESPACE, value)
            case c:
                value = c
                while code[curr + 1] in ALPH_NUMS:
                    curr += 1
                    value += code[curr]
                add_to_tokens(tokens, line_count, curr - start, TokenType.NAME, value)
        curr += 1
    return tokens
